
#ifndef HYPERTEXTURECLASS
#define HYPERTEXTURECLASS

#include <Inventor/SbLinear.h>
#include <stdlib.h>
#include <math.h>

// Perlin noise,
// based on "Hypertextures" by Ken Perlin

class HyperTexture {

  private:

  static const int n=256;
  SbVec3f G[n];
  int P[n];

  public:

  HyperTexture(){
    init();
  }
  
  void init();

  float val(const SbVec3f &x, float b=0.5, float g=0.5) const {
    return bias(gain(turb(x, 10), g), b);
    //return fabs(noise(x/100.0f));
  }

  float val_fractal(const SbVec3f &x, float b=0.5, float g=0.5, float f=1.0) const {
    return gain_0(gain_0(fractal(x, 10, f), g), b);
  }
  
  float bias(float x, float b) const {
    return powf(x, log(b)*(1.0f/log(0.5f)));
  }

  float gain_0(float x, float b) const {
    if (x<0.0)
      return -powf(-x, log(b)*(1.0f/log(0.5f)));
    else
      return powf(x, log(b)*(1.0f/log(0.5f)));
  }
  
  float gain(float x, float g) const {
    if (x<0.5f)
      return (bias(2.0f*x, 1.0f-g)/2.0f);
    else
      return (1.0f - bias(2.0f - 2.0f*x, 1.0f-g)/2.0f);
  }

  float noise(SbVec3f x) const;
  
  float turb(SbVec3f x, int i) const {
    float y = 0.0;
    int a = 1;
    for (int j=0; j<i; j++){
      y += fabs(noise(x*a)/a);
      a <<= 1;
    }

    return y;
  }

  float fractal(SbVec3f x, int i, float f) const {
    float y = 0.0;
    int a = 1;
    for (int j=0; j<i; j++){
      y += noise(x*(a*f))/(a*f);
      a <<= 1;
    }

    return y;
  }

  float Omega(int i, int j, int k, SbVec3f x) const {
    return( omega(x[0])*omega(x[1])*omega(x[2])*Tau(i,j,k,x).dot(x) );
  }

  float omega(float x) const {
    x = fabs(x);
    return( x<1.0f ? (2.0f*x*x*x - 3.0f*x*x + 1.0f) : 0.0f );
  }

  SbVec3f Tau(int i, int j, int k, SbVec3f x) const {
    return G[phi(i+phi(j+phi(k)))];
  }

  int phi(int i) const {
    return P[(i%n+n)%n];
  }

};

#endif /* HYPERTEXTURECLASS */
