//WWWWWW'''''''WW''''''''''''''''WWW'''''''WWWWWW'''''''''''''''''''''''''WWW'
//  WW                            WW         WW                            WW 
//  WW    W WW WW  WWW WWW   WWW  WW   WW    WW    W WW  WW   WW      WWW  WW 
//  WW    WW   WW W WW WW W W WW  WW  WWWW   WW    WW   WWWW WWWW    WW W  WW 
//  WW    WW   WW W WW WW W W WW  WW  WW     WW    WW   WW   WW      WW    WW 
//  WW    WW   WW  WWW WW W  WWW WWWW  WWW   WW    WW    WWW  WWW     WWW WWWW
//                            WW                                              
//''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''''''''''

#ifndef OCTREECLASS
#define OCTREECLASS

#include "Main.h"
#include "Triangle.h"
#include "TriangleList.h"
#include "RayAccel.h"
class Ray; // forward declaration of Ray class

#define EPSILON_OS  0.03    // (Octree Size)
#define EPSILON_OI  0.00005 // (Octree Include)
//#define SHOWTRAV 1


class Octree : public RayAccel {
private:

  // cell
  typedef struct node_st {
    struct node_st *child[8];   // 32 bytes
    double mid[3];              // 24 bytes
    TriangleList *tri_list;     // 4 bytes

#ifdef SHOWTRAV  
    // debugging
    SbVec3f min, max;
    int depth;
#endif
  } node;
  
  // parameters struct
  typedef struct treeParam {
    Ray *r_p;                 // ray
    SbVec3f invDir;
    Triangle *ignoreTri_p;
    //unsigned int rayId;
    unsigned char a;   // used for octree traversal algorithm
  } treeParam;
  
  const static int delta[8][3];
  int octreeDebug;
  int maxItems;
  int maxDepth;
  node *root;
  node *compactNodes;
  int compactNodesSize;
  TriangleList tri_list;
  unsigned int rayId;
  pthread_mutex_t lock;

  // temporary!
  int rays;

//   int num_tris;
//   double maxSideLength;
//   float min[3], max[3];

//   // keep bounding sphere of triangles
//   SbVec3f center;
//   float radius;
//   float shiny_radius;
//   SbVec3f shiny_center;

  // also keep bounding sphere of specular triangles
  SbVec3f shiny_min, shiny_max;
  
  void printStats(node *ptr, int *numCubes = NULL, int *numTris = NULL);
  double findDividingPlane(node *n, int axis, float vmin[3], float vmax[3]);
  void refineNode(node *n, int depth, int simpleDiv,
		  float vmin[3], float vmax[3]);
  inline int new_node(double &t1, int i1, double &t2,
		      int i2, double &t3, int i3) const;
  int first_node(double &tx0, double &ty0, double &tz0,
		 double &txm, double &tym, double &tzm) const;
  int proc_subtree(double &tx0, double &ty0, double &tz0,
		   double &tx1, double &ty1, double &tz1,
		   node *n, treeParam &tp) const;
  int ss_intersect_ray(Ray &r);
  void deleteNode(node *n);

public:

  Octree();
  ~Octree();
  void reset();

  int add(Triangle *t){

    // add triangle
    num_tris++;
    return tri_list.add(t);    
  }

  void printStats();
  void setParams(int maxD, int maxI, int debug);
  void refine(int simpleDiv = 0);
  void reorderNodes();
  int intersect_ray(Ray &r);
  
//   int size(){    return(num_tris);    }
//   double& getMaxSideLength(){ return maxSideLength; }
//   float* getMin(){ return min; }
//   float* getMax(){ return max; }
//   SbVec3f& getCenter(){ return center; }
//   float& getRadius(){ return radius; }
//   float& getShinyRadius(){ return shiny_radius; }
//   SbVec3f& getShinyCenter(){ return shiny_center; }
};

#endif /* OCTREECLASS */
