#ifndef PIXELITERATOR_H
#define PIXELITERATOR_H

#include <sys/time.h>
#include <time.h>
#include "Main.h"

// simple row iterator
class PixelIterator {
protected:
  int imgDim[2];
  int lowerLeft[2], upperRight[2];
  int id, orderId, numThreads;
  int i, j, oldi, oldj;
  float *completion;
  char *completion_stage;
  struct timeval lastTime;
  bool started;
  
public:

  // need to pass in arrays for completion
  PixelIterator(int *imgDim_, int *lowerLeft_, int *upperRight_,
                int id_, int numThreads_,
                float *completion_, char *completion_stage_){
    vcopy2(imgDim, imgDim_);
    vcopy2(lowerLeft, lowerLeft_);
    vcopy2(upperRight, upperRight_);
    orderId = id = id_;
    i = j = 0;
    numThreads = numThreads_;
    completion = completion_;
    completion_stage = completion_stage_;
    lastTime.tv_sec = lastTime.tv_usec = 0;
    started = false;
  }

  virtual bool next(int &x, int &y);

//   virtual void init(){
//     i = lowerLeft[X];
//     j = lowerLeft[Y]+orderId;
//   }
  
//   virtual void reset(){
//     init();
//     fill(completion, completion+numThreads, 0.0); // does this work?
//     fill(completion_stage, completion_stage+numThreads, ' ');
//   }
  
  void setStage(char c=' '){ completion_stage[id]=c; }

  // completion meter
  virtual void printStatus(float val=-1.0);
  virtual int getTileHalfSize() const { return 1; }
  virtual int getTileIteration() const { return 0; }
  virtual void setOrderId(int oId){ orderId = oId; }
};

// progressive iterator
class ProgressiveIterator : public PixelIterator {
private:
  int renderAreaWidth, renderAreaHeight;
  int tileHalfSize, tileSize, tileIteration;
  int firstDrawn, pixelCount;
  int tileI, tileJ, innerI, innerJ;
  
public:
  ProgressiveIterator(int *imgDim_, int *lowerLeft_, int *upperRight_,
                      int id_, int numThreads_,
                      float *completion_, char *completion_stage_) :
    PixelIterator(imgDim_, lowerLeft_, upperRight_, id_, numThreads_,
                  completion_, completion_stage_) {

    // compute next pixel to render
    renderAreaWidth = upperRight[X]-lowerLeft[X];
    renderAreaHeight = upperRight[Y]-lowerLeft[Y];
    tileHalfSize = nearestPowerOf2(MAX2(renderAreaWidth, renderAreaHeight));
    tileSize = tileHalfSize << 1;
    tileIteration = 1;
    firstDrawn = 0;
    pixelCount = 0;
    tileI=tileJ=0;
    innerI=innerJ=0;
    started=false;
  }

  bool next(int &x, int &y);
  void printStatus(float val=-1.0);
  int getTileHalfSize() const { return tileHalfSize; }
  int getTileIteration() const { return tileIteration; }
};


#endif
