#include "PreviewQuad.h"
#include "GlobalVars.h"

PreviewQuad::PreviewQuad(){

  // init
  previewSep = NULL;
  previewTexture = NULL;
  previewSize = SbVec2s(0,0);
  previewRedraw = 0;
  previewCam = NULL;
  previewBufFloat = NULL;
  previewBufByte = NULL;
  previewIter = NULL;
  previewSum = NULL;
  previewCnt = NULL;
  previewDrawnCnt = NULL;
  previewMult = 1.0;
}

PreviewQuad::~PreviewQuad(){
  delete [] previewBufFloat;
  delete [] previewIter;
  delete [] previewSum;
  delete [] previewCnt;
  delete [] previewDrawnCnt;

//  if (previewSep) previewSep->unref();
  //delete previewSep;
  previewSep = NULL;
  
//  if (previewTexture) previewTexture->unref();
  //delete previewTexture;
  previewTexture = NULL;

//  if (previewCam) previewCam->unref();
  //delete previewCam
  previewCam = NULL;
  
  previewBufFloat = NULL;
  previewIter = NULL;
  previewSum = NULL;
  previewCnt = NULL;
  previewDrawnCnt = NULL;
  
  previewSize = SbVec2s(0,0);
  previewRedraw = 0;
  previewBufFloat = NULL;
  previewBufByte = NULL;
  previewMult = 1.0;
}

SoSeparator* PreviewQuad::constructScene(int *imgDim_, int numThreads_){

  vcopy2(imgDim, imgDim_);
  numThreads = numThreads_;
  
  // round up texture size to next power of 2
  previewSize.setValue(imgDim[0], imgDim[1]);
  if (previewSize[0]>0 && previewSize[1]>0){
    previewSize[0] = nearestPowerOf2(MAX2(previewSize[0], previewSize[1]));
    previewSize[1] = previewSize[0];
    // dont make it too big
  }

  // build scene
  // add rendered image plane
  previewSep = new SoSeparator();
  previewSep->ref();

  SoOrthographicCamera *iCam = new SoOrthographicCamera();
  //iCam->viewportMapping.setValue(SoCamera::LEAVE_ALONE);
  iCam->viewportMapping.setValue(SoCamera::ADJUST_CAMERA);
  //iCam->viewportMapping.setValue(SoCamera::CROP_VIEWPORT_NO_FRAME);
  //iCam->viewportMapping.setValue(SoCamera::CROP_VIEWPORT_LINE_FRAME);
  previewSep->addChild(iCam);
  previewCam = iCam;

  SoResetTransform *iReset = new SoResetTransform();
  iReset->whatToReset.setValue(SoResetTransform::BBOX);
  previewSep->addChild(iReset);

  SoLightModel *iLightModel = new SoLightModel();
  iLightModel->model.setValue(SoLightModel::BASE_COLOR);
  previewSep->addChild(iLightModel);

  SoTransform *iTran = new SoTransform();
  float aspect = imgDim[X]/(float)imgDim[Y];
  if (aspect>=1.0) iTran->scaleFactor.setValue(aspect, 1, 1);
  else             iTran->scaleFactor.setValue(1, 1/aspect, 1);
  previewSep->addChild(iTran);
  
  SoCoordinate3 *iCoord = new SoCoordinate3();
  iCoord->point.set1Value(0, SbVec3f(-1, -1, 0));
  iCoord->point.set1Value(1, SbVec3f( 1, -1, 0));
  iCoord->point.set1Value(2, SbVec3f( 1,  1, 0));
  iCoord->point.set1Value(3, SbVec3f(-1,  1, 0));
  previewSep->addChild(iCoord);

  SoComplexity *iComp = new SoComplexity();
  iComp->value.setValue(0.0f);
  iComp->textureQuality.setValue(0.8f);
  previewSep->addChild(iComp);

  SbVec2s texSize = previewSize;

  float iTexCoord[2] = {
    imgDim[0]/(float)texSize[0],
    imgDim[1]/(float)texSize[1]
  };
  
  SoTextureCoordinate2 *iTextCoord = new SoTextureCoordinate2();
  iTextCoord->point.set1Value(0, SbVec2f(0, 0));
  iTextCoord->point.set1Value(1, SbVec2f(iTexCoord[0], 0));
  iTextCoord->point.set1Value(2, SbVec2f(iTexCoord[0], iTexCoord[1]));
  iTextCoord->point.set1Value(3, SbVec2f(0, iTexCoord[1]));
  previewSep->addChild(iTextCoord);

  SoTexture2 *iTex = new SoTexture2();
  iTex->model.setValue(SoTexture2::MODULATE);
  previewSep->addChild(iTex);
  previewTexture = iTex;

  // init texture
  int nc = 3;
  int texVals = texSize[0]*texSize[1]*nc;
  unsigned char *imgbuf = new unsigned char[texVals];
  for (int i=0; i<texVals; i++) imgbuf[i] = 128;
  iTex->image.setValue(texSize, nc, imgbuf);
  delete [] imgbuf;
  previewBufByte = previewTexture->image.startEditing(texSize, nc);

  SoMaterial *iMat = new SoMaterial();
  iMat->diffuseColor.setValue(1,1,1);
  iMat->transparency.setValue(0);
  previewSep->addChild(iMat);

  SoIndexedFaceSet *iFaceSet = new SoIndexedFaceSet();
  iFaceSet->coordIndex.set1Value(0, 0);
  iFaceSet->coordIndex.set1Value(1, 1);
  iFaceSet->coordIndex.set1Value(2, 2);
  iFaceSet->coordIndex.set1Value(3, 3);
  iFaceSet->coordIndex.set1Value(4, -1);
  previewSep->addChild(iFaceSet);

  constructed=true;
  return previewSep;
}

void PreviewQuad::preRender(){

  if (!constructed){
    cerr << "ERROR: PreviewQuad::preRender(): You should have "
      "called constructScene() first.\n";
    return;
  }
  
  // init buffers
  int size = previewSize[X]*previewSize[Y];
    
  if (!previewBufFloat) previewBufFloat = new SbColor[size];
  if (!previewIter) previewIter = new unsigned char[size];
  if (!previewSum) previewSum = new float[numThreads];
  if (!previewCnt) previewCnt = new int[numThreads];
  if (!previewDrawnCnt) previewDrawnCnt = new int[numThreads];
  previewMult = 1.0f;    
    
  for (int i=0; i<size; i++){
    previewBufFloat[i] = SbColor(0,0,0);
    previewBufByte[i] = 128;
    previewIter[i] = 0;
  }
  for (int i=0; i<numThreads; i++){
    previewSum[i] = 0.0f;
    previewCnt[i] = 0;
    previewDrawnCnt[i] = 0;
  }

  enabled=true;
}

// callback function to redraw the preview
// (frame buf -> float buf -> byte buf -> *texture map*)
void redrawSceneCB(void *data, SoSensor *sensor){
  GlobalVars *gv = (GlobalVars *)data;
  static int redraws = 1;
  if (gv->preview->getRedraw()){
    gv->preview->setRedraw(false);
    gv->preview->finishTexture();
    cerr << "R" << redraws++ << " ";
    gv->xtViewer->render();
  }
}

// enables the preview display
// the "master switch" that switches the scene graph
// and turns on a timer sensor to redraw the preview texture
void PreviewQuad::togglePreviewDisplay(GlobalVars *gv, bool turnOn){
  static SoTimerSensor *sensor = NULL;
  static bool previewOn = false;

  if (sensor == NULL)
    sensor = new SoTimerSensor(redrawSceneCB, gv);

  int curval = gv->sceneSwitch->whichChild.getValue();

  extern void toggleSceneSwitch(GlobalVars *gv, int val1=-2, int val2=-2);
  
  // turn on
  if (turnOn || previewOn == false){
    sensor->setInterval(SbTime(0.5)); 
    sensor->schedule();
    previewOn = true;
    toggleSceneSwitch(gv, 1, 1); // set on
    cerr << "Preview on.\n";
  }

  // turn off
  else {
    sensor->unschedule();
    previewOn = false;
    if (curval == 1)
      toggleSceneSwitch(gv); // old val
    cerr << "Preview off.\n";
  }
}


void PreviewQuad::redrawPreview(const int *lowerLeft,
			 const int *upperRight){

  if (!enabled) return;
  
  const float mult = 255.0f*previewMult;
  for (int j=lowerLeft[Y]; j<upperRight[Y]; j++){
    for (int i=lowerLeft[X]; i<upperRight[X]; i++){
      int idx = previewSize[X]*j+i;
      unsigned int r,g,b;
      r = clamp((int)(previewBufFloat[idx][0]*mult), 0, 255);
      g = clamp((int)(previewBufFloat[idx][1]*mult), 0, 255);
      b = clamp((int)(previewBufFloat[idx][2]*mult), 0, 255);
      previewBufByte[idx*3+0] = r;
      previewBufByte[idx*3+1] = g;
      previewBufByte[idx*3+2] = b;
    }
  }

  // schedule redraw
  previewRedraw = 1;
}
  
void PreviewQuad::updatePreview(const int *lowerLeft,
			 const int *upperRight){

  if (!enabled) return;
  
  // compute new scale factor
  float avg = 0;
  int threads=0;
  for (int i=0; i<numThreads; i++){
    if (!previewCnt[i])
      continue;
    avg += previewSum[i]/previewCnt[i];
    threads++;
  }
  if (threads)
    avg /= threads;
  if (avg)
    previewMult = 0.5f/avg;
  cerr << "previewMult=" << previewMult << "\n";

  redrawPreview(lowerLeft, upperRight);
}

void PreviewQuad::copyPixToPreview(int *lowerLeft, int *upperRight,
                                   SbColor *pix){
  
  if (enabled) return;

  for (int j=lowerLeft[Y]; j<upperRight[Y]; j++){
    for (int i=lowerLeft[X]; i<upperRight[X]; i++){
      int idx = previewSize[X]*j+i;
      previewBufFloat[idx] = pix[j*imgDim[X]+i];
    }
  }

}

void PreviewQuad::setPreviewPix(int x, int y, int tId, const SbColor &col,
                         int size, int iter,
                         int *lowerLeft, int *upperRight){

  if (!enabled) return;

  previewSum[tId] += (col[0]+col[1]+col[2])*(1/3.f);
  previewCnt[tId]++;
    
  unsigned int r,g,b;
  float mult = 255.0f*previewMult;
  r = clamp((int)(col[0]*mult), 0, 255);
  g = clamp((int)(col[1]*mult), 0, 255);
  b = clamp((int)(col[2]*mult), 0, 255);

  int min[2], max[2];
  min[0] = MAX2(x, lowerLeft[0]);
  min[1] = MAX2(y, lowerLeft[1]);
  max[0] = MIN2(x+size, upperRight[0]);
  max[1] = MIN2(y+size, upperRight[1]);
  
  for (int j=min[1]; j<max[1]; j++){
    for (int i=min[0]; i<max[0]; i++){
      int idx = previewSize[X]*j+i;
      if (previewIter[idx]>iter) continue;
      previewIter[idx] = iter;
      previewBufFloat[idx] = col;
      previewBufByte[idx*3+0] = r;
      previewBufByte[idx*3+1] = g;
      previewBufByte[idx*3+2] = b;
    }
  }

  previewDrawnCnt[tId] += (max[0]-min[0])*(max[1]-min[1]);

  float pix = MAX2(256,(upperRight[0]-lowerLeft[0])*(upperRight[1]-lowerLeft[1])/10);
  if (tId==0 && previewDrawnCnt[tId]*numThreads>pix){
    previewRedraw = 1;
    previewDrawnCnt[tId] = 0;
  }

  // update preview after (nearly full) redraw
  if (y==lowerLeft[1] && x==(lowerLeft[0]+size))
    updatePreview(lowerLeft, upperRight);
}
