#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include "SoMatPlus.h"
#include "debug.h"
#include <iostream>

SO_NODE_SOURCE(SoMatPlus);

void SoMatPlus::initClass(){
  SO_NODE_INIT_CLASS(SoMatPlus, SoNode, "Node");
}

// constructor
SoMatPlus::SoMatPlus(){
  SO_NODE_CONSTRUCTOR(SoMatPlus);

  SO_NODE_ADD_FIELD(model, (SCHLICK));  
  SO_NODE_ADD_FIELD(rough1, (1.0));
  SO_NODE_ADD_FIELD(rough2, (1.0));
  SO_NODE_ADD_FIELD(sym1, (1.0));
  SO_NODE_ADD_FIELD(sym2, (1.0));
  SO_NODE_ADD_FIELD(R0, (1.0));
  SO_NODE_ADD_FIELD(layers, (1));

  /*
  SoSFFloat rough1, rough2;  // roughness  (0=mirror, 1=diffuse (default))
  SoSFFloat sym1, sym2;      // anisotropy
                             // (0=aniisotropic, 1=isotropic (symmetrical, default))
  SoSFFloat R0;              // reflection at normal incidence
                             // Default: 1
                             // -1 means auto-alculated based on index of refractions

  SoSFInt layers;            // 1 (default) or 2
  //SoSFVec2f g, s, d;         // internal (private, automatically calculated)
  */

  //SO_NODE_ADD_FIELD(transparency, (0.0));
  SO_NODE_ADD_FIELD(ior1, (1.0));
  SO_NODE_ADD_FIELD(ior2, (1.5));
  SO_NODE_ADD_FIELD(atten, (1.0, 1.0, 1.0));

  SO_NODE_ADD_FIELD(emisEcc, (1.0));
  SO_NODE_ADD_FIELD(minDist, (0.0001));

  SO_NODE_DEFINE_ENUM_VALUE(ReflModel, SCHLICK);
  SO_NODE_DEFINE_ENUM_VALUE(ReflModel, DATA);
  SO_NODE_DEFINE_ENUM_VALUE(ReflModel, LIBRARY);
  SO_NODE_DEFINE_ENUM_VALUE(ReflModel, BSSRDF);
  SO_NODE_DEFINE_ENUM_VALUE(ReflModel, HIERSUBSURF);
  
  SO_NODE_SET_SF_ENUM_TYPE(model, ReflModel);
  
  /*
  float transparency; // 0 (opaque, default) - 1 (clear)
  SoSFFloat ior1;     // OUTSIDE index of refraction Default: 1.0 (air)
  SoSFFloat ior2;     // INSIDE index of refraction Default: 1.5 (glass)
  SoSFColor atten;    // attenuation (fraction) at 1m when going through inside
                      // Default (0.98, 0.99, 0.98)
                      // from specularColor ?
  */

  isBuiltIn = TRUE;   // pertains to whether .iv files need extra [...] field
}

SoMatPlus::~SoMatPlus(){
}

void SoMatPlus::GLRender(SoGLRenderAction *action){
//  doAction(action, "GLRender");
}

void SoMatPlus::callback(SoCallbackAction *action){
  doAction(action, "callback");
}

void SoMatPlus::doAction(SoAction *action, const char *s){
  cerr << "hi from SoMatPlus::doAction(" << s << ")\n";
}

// ostream& operator<< (ostream& os, const SbVec3f& x){
//   os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
//   os << "(" << x[X] << ", " << x[Y] << ", " << x[Z] << ")";
//   resetios(os);
//   return(os);
// }

int matplus_main(int argc, char *argv[]){
//int main(int argc, char *argv[]){
  SoMatPlus *hi[50];
  for (int i=0; i<50; i++)
    hi[i] = new SoMatPlus;
  for (int i=0; i<50; i++)
    hi[i]->ref();
  for (int i=0; i<50; i++)
    hi[i]->unref();
  return(0);
}

ostream& operator<< (ostream& os, const SoMatPlus & m){
  os << "MatPlus ["
     << " model=";

  switch (m.model.getValue()){
  case SoMatPlus::SCHLICK:
    os << "schlick";
    break;
  case SoMatPlus::DATA:
    os << "data";
    break;
  case SoMatPlus::LIBRARY:
    os << "library";
    break;
  case SoMatPlus::BSSRDF:
    os << "bssrdf";
    break;
  case SoMatPlus::HIERSUBSURF:
    os << "hiersubsurf";
    break;
  }
  
  os << " rough1=" << m.rough1.getValue()
     << " rough2=" << m.rough2.getValue()
     << " R0=" << m.R0.getValue()
     << " layers=" << m.layers.getValue()
     << " ior1=" << m.ior1.getValue()
     << " ior2=" << m.ior2.getValue()
     << " atten=" << m.atten.getValue()
     << " emisEcc=" << m.emisEcc.getValue()
     << " minDist=" << m.minDist.getValue()
     << " ]";
  
  return(os);
}
