//WWWWWW'''''''''''''WW'''''''''''''''''''''''''WWW''''''''''''''''''
//  WW               WW                          WW                  
//  WW     WW  WW W WWWW W WW W WW  WW      WWW  WW   WWW  WWWW  WWWW
//  WW    WWWW  WW   WW  W WW WW   WWWW    WW W  WW  W WW WWW   WWW  
//  WW    WW    WW   WW  W WW WW   WW      WW    WW  W WW   WWW   WWW
//  WW     WWW W WW  WW   WWW WW    WWW     WWW WWWW  WWW WWWW  WWWW 
//                                                                   
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#include "Texture.h"
#include "debug.h"

Texture::Texture(){
  wrapS = SoTexture2::REPEAT;
  wrapT = SoTexture2::REPEAT;
  model = SoTexture2::MODULATE;
  numComps = 3;
  bumpMap = 0;
  bumpScale = 1.0;
  img = NULL;
  imgTexture = NULL;
  size = SbVec2s(0, 0);
  blend = SbColor(0.0, 0.0, 0.0);
  stex[0] = ttex[0] = 0.0;
  stex[1] = 1.0; ttex[1] = 0.0;
  stex[2] = 0.5; ttex[2] = 1.0;
}

Texture::~Texture(){
  if (texture_maps.count(img) != 0){
    delete [] texture_maps[img];
    texture_maps.erase(img);
  }
}

Texture::Texture(int s, int t, int m,
		 SbColor b, int n,  SbVec2s &siz,
		 const unsigned char *_img,
		 int _bumpMap, float _bumpScale){
  wrapS = s;
  wrapT = t;
  model = m;
  blend = b;
  numComps = n;
  size = siz;
  img = _img;
  bumpMap = _bumpMap;
  bumpScale = _bumpScale;
  addTextureMap(img, numComps, size, imgTexture);

  // stex[], ttex[] get filled in by Triangle(...)
}


SbColor Texture::getColor(const unsigned char *img, int numComps,
			  float s, float t,
			  SbColor &color){

  float r, g, b;
  float i_f, j_f;
  float sp, tp;
  int i, j, i1, j1;
  
  // do wrapping
  if (wrapS == SoTexture2::REPEAT){
    s -= (float)((int)s);
    if (s<0.0)
      s += 1.0;

    t -= (float)((int)t);
    if (t<0.0)
      t += 1.0;
  }
  // or clamp
  else {
    s = clamp(s, 0.0f, 1.0f);
    t = clamp(t, 0.0f, 1.0f);
  }
  
  // calculate indices
  i  = (int)floor(s*size[0]);
  j  = (int)floor(t*size[1]);
  i1 = i+1;
  j1 = j+1;

  if (wrapS == SoTexture2::REPEAT){
    if (i1>size[0]-1)
      i1 = 0;
    if (j1>size[1]-1)
      j1 = 0;
  }
  else {
    i1 = clamp(i1, 0, size[0]-1);
    j1 = clamp(j1, 0, size[1]-1);
  }
  
//  i = clamp(i, 0, size[0]-1);
//  j = clamp(j, 0, size[1]-1);
 
  // calculate texel ratios
  sp = s*size[0]-i;
  tp = t*size[1]-j;
	 
  // perform hermite smoothing
  sp = 3.0*sp*sp-2.0*sp*sp*sp;
  tp = 3.0*tp*tp-2.0*tp*tp*tp;

  if (numComps == 3){
	 
    r = (1.0-sp)*(1.0-tp)*img[size[0]*3*j +i *3+R] +
              sp*(1.0-tp)*img[size[0]*3*j +i1*3+R] +
        (1.0-sp)*tp      *img[size[0]*3*j1+i *3+R] +
              sp*tp      *img[size[0]*3*j1+i1*3+R];

    g = (1.0-sp)*(1.0-tp)*img[size[0]*3*j +i *3+G] +
             sp *(1.0-tp)*img[size[0]*3*j +i1*3+G] +
        (1.0-sp)*tp      *img[size[0]*3*j1+i *3+G] +
              sp*tp      *img[size[0]*3*j1+i1*3+G];

    b = (1.0-sp)*(1.0-tp)*img[size[0]*3*j +i *3+B] +
              sp*(1.0-tp)*img[size[0]*3*j +i1*3+B] +
        (1.0-sp)*tp      *img[size[0]*3*j1+i *3+B] +
              sp*tp      *img[size[0]*3*j1+i1*3+B];

    color = SbColor(r/255.0, g/255.0, b/255.0);
  }
  else if (numComps == 1){

    // debugging
    //cerr << "numComps = 1!!! \n";
    
    r = (1.0-sp)*(1.0-tp)*img[size[0]*j +i ] +
      sp*(1.0-tp)      *img[size[0]*j +i1] +
      (1.0-sp)*tp      *img[size[0]*j1+i ] +
      sp*tp            *img[size[0]*j1+i1];
      
    color = SbColor(r/255.0, r/255.0, r/255.0);
  }

  return (color);
}


// texture filtering from "Realistic Ray Tracing" by Peter Shirley p.61
SbColor Texture::getColor(const SbVec3f &bc, const SbColor diffuse[3]){
  SbColor color;
  float s, t;
    
  // calculate image coordinates (interpolate)
  s = bc[0]*stex[0] + bc[1]*stex[1] + bc[2]*stex[2];
  t = bc[0]*ttex[0] + bc[1]*ttex[1] + bc[2]*ttex[2];

  //return(SbColor(s,t,0.0));
  getColor(imgTexture, numComps, s, t, color);

  // blend model
  // multiply components by interpolated diffuse color
  if (model == SoTexture2::MODULATE){

    // this uses the same diffuse color barycentric interpolation as in
    // Triangle::getIntDiffColor()
    color[0] *= bc[0]*diffuse[0][R] + bc[1]*diffuse[1][R] + bc[2]*diffuse[2][R];
    color[1] *= bc[0]*diffuse[0][G] + bc[1]*diffuse[1][G] + bc[2]*diffuse[2][G];
    color[2] *= bc[0]*diffuse[0][B] + bc[1]*diffuse[1][B] + bc[2]*diffuse[2][B];
  }

  // use texture as blending factor between diffuseColor and blend color
  // white texture = 100% blend color, black = 100% diffuseColor
  else if (model == SoTexture2::BLEND){

    SbColor diffCol = SbColor(
      bc[0]*diffuse[0][R] + bc[1]*diffuse[1][R] + bc[2]*diffuse[2][R],
      bc[0]*diffuse[0][G] + bc[1]*diffuse[1][G] + bc[2]*diffuse[2][G],
      bc[0]*diffuse[0][B] + bc[1]*diffuse[1][B] + bc[2]*diffuse[2][B]
      );

    float intensity = (color[0]+color[1]+color[2])/3.0;
    color = intensity*blend + (1.0-intensity)*diffCol;
  }
    
  return(color);
}

typedef map<const unsigned char *, unsigned char *> PointerMap;
PointerMap Texture::texture_maps;

void Texture::addTextureMap(const unsigned char *img,
			    int numComps, SbVec2s size,
			    unsigned char *&imgTexture){
  
  if (texture_maps.count(img) == 0){
      // process rgb triplets into reflectances
      int bytes = size[0]*size[1]*numComps;
      unsigned char *img2 = new unsigned char[bytes];
      if (img2 == NULL){
	cerr << "Error: Could not allocate memory ("
	     << bytes << " bytes) for texture map!\n";
	exit(1);
      }
//      cerr << "Correcting " << bytes << " bytes.\n";
      for (int i=0; i<bytes; i++){
	float x = (float)img[i]/255.0;     // convert to float
//	float d = -1.45*x+1.5;       // convert to optical density
//	float r = pow(10.0f, -d);     // convert to reflectance

	// this funciton is a polynomial fit of w, z
	// w = [ 0 .2 .4 .6 .8 1 ]           (evenly spaced 0..1)
	// x = [ 0 53 117 180 249 255 ]/255  (RGB of macbeth b&w swatches)
	// y = [ 1.5 1.05 .7 .44 .23 .05 ]   (optical density of swatches)
	// z = 10^-y                         (10.^-y in matlab)
	//
	// E.g. (in matlab):
	//
	// w=0:1/5:1
	// y=[1.5 1.05 .7 .44 .23 .05]
	// z=10.^-y
	// [P,S]=polyfit(w,z,3)
	// zfit=polyval(P,w)
	//
	float r=0.1689*x*x*x + 0.4994*x*x +0.1914*x + 0.0311;
	r = clamp(r, 0.0f, 1.0f);
	img2[i] = (unsigned char)(r*255.0);
//	cerr << "texture: " << int(img[i])
//	     << " -> " << int(img2[i]) << "\n";
//	img2[i] = img[i];
      }
      texture_maps[img] = img2;
//      cerr << "Saving tm[" << int(img) << "] = " << int(img2) << endl;

      /*
  // debugging
  cout << "\nTexture:\n";
  for (int j=0; j<size[1]; j++){
    for (int i=0; i<size[0]; i++){
      int c =img2[size[0]*3*j +i *3+R];
      float cf = c/256.0*10.0;
      char cc = (int)cf+'0';
      cout << cc;
    }
    cout << "\n";
  }
  cout << "\n";
      */
  }
  else {
//    cerr << "Loading tm[" << int(img) << "] = "
//	 << int(texture_maps[img]) << endl;
  }

  imgTexture = texture_maps[img];
}


SbVec3f Texture::getDeriv(const SbVec3f &bc){

  float s, t;
  float ds, dt;

  if (!bumpMap)
    return(SbVec3f(0.0, 1.0, 0.0));

  // calculate image coordinates (interpolate)
  s = bc[0]*stex[0] + bc[1]*stex[1] + bc[2]*stex[2];
  t = bc[0]*ttex[0] + bc[1]*ttex[1] + bc[2]*ttex[2];

  ds = 1.0/size[0];
  dt = 1.0/size[1];
  
  SbColor color1, color2, color3;
  float b1, b2, b3;
  getColor(imgTexture, 3, s,    t,    color1);
  getColor(imgTexture, 3, s+ds, t,    color2);
  getColor(imgTexture, 3, s,    t+dt, color3);

  b1 = brightness(color1);
  b2 = brightness(color2);
  b3 = brightness(color3);
  
  SbVec3f deriv;
  deriv[R] = (b2-b1)/ds;
  deriv[G] = (b3-b1)/dt;
  deriv[B] = b1;
  
  return(deriv);
}
  
