//WWWWWW'''''''WW''''''''''''''''WWW'''''''''''''''WWW''''''''''''''''''
//  WW                            WW                WW                  
//  WW    W WW WW  WWW WWW   WWW  WW   WW      WWW  WW   WWW  WWWW  WWWW
//  WW    WW   WW W WW WW W W WW  WW  WWWW    WW W  WW  W WW WWW   WWW  
//  WW    WW   WW W WW WW W W WW  WW  WW      WW    WW  W WW   WWW   WWW
//  WW    WW   WW  WWW WW W  WWW WWWW  WWW     WWW WWWW  WWW WWWW  WWWW 
//                            WW                                        
//''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''''

#ifndef TRIANGLECLASS
#define TRIANGLECLASS

#include "Main.h"
#include "Misc.h"
#include "Material.h"
class Texture; // forward declaration of Texture class

class Triangle {
protected:
  static int errorTris;
  static float emissiveArea;

//  static vector<SbVec3f
    
public:
  // geometry 
  SbVec3f vert[3];
  SbVec3f normals[3];
//  SbVec3f normal;           // Not really needed (except for lights)

  // material
  Material *mat;
  SbColor diffuse[3];
//  SbColor specular;         // Could be stored with shader
//  SbColor emission;         // Could be stored with shader (only needed for lights)
  // SbShader *shader;

  // candidates for removal
  Texture *texture;         // Texture object is bloated
//  float shininess, transparency;  // Could be stored in shader
//  unsigned char shiny, emissive;  // Could be stored in shader
  
//  char shown;   // Not necessary
  
  
  Triangle();
  Triangle(void *data, SoCallbackAction *action,
	   const SoPrimitiveVertex *v0,
	   const SoPrimitiveVertex *v1,
	   const SoPrimitiveVertex *v2);
  ~Triangle();

  void calcBaryCoords(const SbVec3f &pnt, SbVec3f &bc) const;
  void useBaryCoords(const double a, const double b, const double c,
		     SbVec3f &bc);
  
  SbColor& getSpecular()         const { return(mat->specular);     }
  int      isShiny()             const { return(mat->shiny);        }
  int      isEmissive()          const { return(mat->emissive);     }
  float    getShininess()        const { return(mat->shininess);    }
  float    getTransparency()     const { return(mat->transparency); }
  virtual  float   getEmisAvg()  const { return(mat->emisAvg);  }
  virtual  SbColor getEmission() const { return(mat->emission); }
  virtual  SbVec3f getNormal()   const;
  SbColor  getIntEmission( const SbVec3f &bc) const;
  SbVec3f  getIntNormal(   const SbVec3f &bc) const;
  SbColor  getIntDiffColor(const SbVec3f &bc) const;

  float    getIndexOfRefractionTop() const { return(mat->matPlus->ior1.getValue());  }
  float    getIndexOfRefractionBot() const { return(mat->matPlus->ior2.getValue());  }
  const SbColor& getAttenuation()    const { return(mat->matPlus->atten.getValue()); }
  
  // draw verts (spheres) + tri (polygon)
  void show(SbColor c = SbColor(1.0, 1.0, 1.0));

  // draw as edge lines, slightly inset
  void show2();

  // display statistical information
  static void printGlobalStats();

};

ostream& operator<< (ostream& os, const Triangle& x);

class EmissiveTriangle : public Triangle {

  private:

  // projection map stuff
  char *map;                 // value of NULL means don't use map
  float *celluv;             // precomputed u,v values of filled (only) cells
  float mapDeltaCDF;         // delta per filled box in map CDF
  float mapPDF;              // distribution of filled boxes
  ONB onb;                   // ortho-normal basis of center of triangle
  static const int mapSize[2];
  int cells;

  public:

  EmissiveTriangle();
  EmissiveTriangle(void *data, SoCallbackAction *action,
		   const SoPrimitiveVertex *v0,
		   const SoPrimitiveVertex *v1,
		   const SoPrimitiveVertex *v2);
  ~EmissiveTriangle(){
    delete [] map;
    delete [] celluv;
  }
  int UseMap(){ return map?true:false; }
  
  // precomputed to accelerate direct lighting calculation
  // or only needed for lights
  float area;
  float emisAvg;
  SbVec3f normal;
  SbColor emission;         // Could be stored with shader (only needed for lights)
  
  float    getEmisAvg()   const { return(emisAvg);  }
  SbColor  getEmission()  const { return(emission); }
  SbVec3f  getNormal()    const { return(normal);   }
  const SbVec3f& getNormalPre() const { return(normal);   } // precomputed normal

  void genImpMap(const SbVec3f &center, float rad);
  void sampleDir(SbVec3f &dir, float &pdf, float e1, float e2);
  void printMap();
  
};

ostream& operator<< (ostream& os, const EmissiveTriangle& x);


#endif /* TRIANGLECLASS */
